; 7-ade configuration
; ©2016–2020 Daniel Beardsmore
; https://telcontar.net/Software/7-ade

get7ZipPath() {
	; Locate the installation of 7-Zip on this computer and return the path containing it
	; More detection methods may need to be added in future :)
		
	sevenZipPath := ""
	
	if (A_Is64bitOS)
		SetRegView, 64
	
	RegRead, sevenZipPath, HKEY_CURRENT_USER, SOFTWARE\7-Zip, Path
	if (!sevenZipPath)
	{
		RegRead, sevenZipPath, HKEY_LOCAL_MACHINE, SOFTWARE\7-Zip, Path
	}
	
	if (sevenZipPath && substr(sevenZipPath, -1) == "\")
		sevenZipPath := substr(sevenZipPath, 0, -1)
	
	if (A_Is64bitOS)
		SetRegView, Default
	
	return sevenZipPath
}

getConfig() {
	; Read the user’s configuration
	; This can live either in the same directory as 7-Up, or in Appdata\Telcontar
	
	; This function in particular needs to do various things that the insane syntax won't allow
	
	; Note that calls to dbg() only work when the 7-ade.ahk is run, as this enforces debug output regardless of whether debugging is enabled
	; within the configuration
	
	configPath := getConfigPath()
	if (!configPath)
		return {}
	
	configFields := { "_stupid": "S"
		,	"use demo": "B"	; Automatically expand demo file for testing purposes
		,	"demo file": "S"	; Name of demo file
		,	"rules": "S"		; Read the user’s rulesets
		,	"debug": "B"		; Write debugging output to DebugView
		,	"_stupid_": "S" }
	ruleFields := { "_stupid": "S"
		,	"name begins": "S", "name ends": "S", "name contains": "S", "name is": "S"
		,	"path begins": "S", "path contains": "S", "path is": "S"
		,	"overwrite": "B", "target path": "S"
		,	"show": "E"
		,	"launch output": "E"
		,	"delete archive": "B"
		,	"recycle archive": "B"
		,	"_stupid_": "S" }
	
	config := readINIsection(configPath, k_APPNAME, configFields)
	config["rulesets"] := []
	
	; v1.0 rules
	dbg("Reading rules")
	if (config["rules"]) {
		stupid := config["rules"]
		Loop, Parse, stupid, `,, %A_Space%%A_Tab%
		{
			rule := readINIsection(configPath, A_LoopField, ruleFields)
			if (rule) {
				rule["name"] := A_LoopField
				config["rulesets"].push(rule)
				dbg("Found v1.0 rule " . A_LoopField)
			}
		}
	}
	
	; v1.1 rules
	IniRead, iniSectionNames, %configPath%
	iniSectionNames := StrSplit(iniSectionNames, "`n")  
	for k, v in iniSectionNames {
		if (substr(v, 1, 6) == "Rules\" && strlen(v) >= 7) {
			rule := readINIsection(configPath, v, ruleFields)
			rule["name"] := substr(v, 7)
			config["rulesets"].push(rule)
			dbg("Found v1.1 rule " . rule["name"])
		}
	}
	
	; Read defaults
	rule := readINIsection(configPath, "Defaults", ruleFields)
	if (rule) {
		rule["name"] := "Defaults"
		config["defaults"] := rule
	}
	
	return config
}

getConfigPath() {
	; Find where the program’s configuration lives
	; Returns false if no INI file exists
	
	configName := k_APPNAME . ".ini"
	configPath := A_ScriptDir . "\" . configName
	
	if (FileExist(configPath))
		return configPath
	
	configPath := A_AppData . "\Telcontar\" . configName
	if (FileExist(configPath))
		return configPath
	
	return false
}


readINIsection(iniPath, section, fields) {
	dict := {}
	
	iniOut := ""
	for fieldName, fieldType in fields
	{
		IniRead, iniOut, %iniPath%, %section%, %fieldName%, %A_Space%
		if (iniOut)
		{
			if (fieldType == "B") {
				iniOut := getHumanBool(iniOut)
			}
			else if (fieldType == "E") {
				; enum type: trim and convert to uppercase for comparison
				iniOut := trim(iniOut)
				StringUpper, iniOut, iniOut
			}
			dict[fieldName] := iniOut
		}
	}
	
	return dict
}


getHumanBool(humanBool) {
	; Convert a human-readable boolean value to true/false -- this allows some flexibility in the INI file contents
	
	humanBool := trim(humanBool)
	StringUpper, humanBool, humanBool
	if (humanBool == "TRUE" || humanBool == "YES" || humanBool == "1" || humanBool == "AFFIRMATIVE, MASTER")
		return true
	
	return false
}